# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-12-20 11:51
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings
import django.db.models.deletion

from django_multitenant.db import migrations as tenant_migrations


def get_operations():
    operations = [
        migrations.RunSQL("ALTER TABLE tests_tenantnotidmodel DROP CONSTRAINT tests_tenantnotidmodel_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_somerelatedmodel DROP CONSTRAINT tests_somerelatedmodel_pkey CASCADE;"),
    ]

    if settings.USE_CITUS:
        operations += [
            tenant_migrations.Distribute('TenantNotIdModel'),
            tenant_migrations.Distribute('SomeRelatedModel'),
        ]

    operations += [
        migrations.RunSQL("ALTER TABLE tests_somerelatedmodel ADD CONSTRAINT tests_somerelatedmodel_pkey PRIMARY KEY (related_tenant_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_tenantnotidmodel ADD CONSTRAINT tests_tenantnotidmodel_pkey PRIMARY KEY (tenant_column);")
    ]

    return operations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0002_distribute'),
    ]

    operations = get_operations()



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-04-12 08:31
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields
import django_multitenant.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0007_alter_foreignkey_field'),
    ]

    operations = [
        migrations.AddField(
            model_name='somerelatedmodel',
            name='opened',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 3.0 on 2020-02-11 10:00

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0017_auto_20200204_0929'),
        ('tests', '0021_auto_20200129_0453'),
    ]

    operations = [
    ]



from django.db import connections, transaction
from django.db.models import Q
from django.db.models.sql.constants import (
    GET_ITERATOR_CHUNK_SIZE, NO_RESULTS,
)
from django.conf import settings


from .utils import (get_current_tenant, get_tenant_column,
                    get_tenant_filters, is_distributed_model)


def add_tenant_filters_on_query(obj):
    current_tenant = get_current_tenant()

    if current_tenant:
        try:
            filters = get_tenant_filters(obj.model)
            obj.add_q(Q(
                **filters
            ))
        except ValueError:
            pass


def wrap_get_compiler(base_get_compiler):
    def get_compiler(obj, *args, **kwargs):
        add_tenant_filters_on_query(obj)
        return base_get_compiler(obj, *args, **kwargs)

    get_compiler._sign = "get_compiler django-multitenant"
    return get_compiler


def wrap_update_batch(base_update_batch):
    def update_batch(obj, pk_list, values, using):
        obj.add_update_values(values)
        for offset in range(0, len(pk_list), GET_ITERATOR_CHUNK_SIZE):
            obj.where = obj.where_class()
            obj.add_q(Q(pk__in=pk_list[offset: offset + GET_ITERATOR_CHUNK_SIZE]))
            add_tenant_filters_on_query(obj)
            obj.get_compiler(using).execute_sql(NO_RESULTS)

    update_batch._sign = "update_batch django-multitenant"
    return update_batch



def wrap_delete(base_delete):
    def delete(obj):
        obj_are_distributed = [is_distributed_model(instance) for instance in obj.data.keys()]

        # If all elements are from distributed tables, then we can do a simple atomic transaction.
        # If there are mixes of distributed and reference tables, it would raise the following error :

        # django.db.utils.InternalError: cannot execute DML on reference relation
        # "table" because there was a parallel DML access to distributed relation
        # "table2" in the same transaction

        if (len(set(obj_are_distributed)) > 1
            and getattr(settings, 'CITUS_EXTENSION_INSTALLED', False)):
            # all elements are not the same, some False, some True
            with transaction.atomic(using=obj.using, savepoint=False):
                connections[obj.using].cursor().execute("SET LOCAL citus.multi_shard_modify_mode TO 'sequential';")
                result = base_delete(obj)
                connections[obj.using].cursor().execute("SET LOCAL citus.multi_shard_modify_mode TO 'parallel';")
                return result

        return base_delete(obj)

    return delete



# Generated by Django 2.2.10 on 2020-05-26 04:42

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0023_auto_20200412_0603'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectLog',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('content', models.CharField(max_length=255)),
            ],
        ),
        migrations.AddField(
            model_name='project',
            name='log',
            field=models.ForeignKey(blank=True, null=True, db_constraint=False, on_delete=django.db.models.deletion.CASCADE, to='tests.ProjectLog'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-11-14 17:28
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Account',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('domain', models.CharField(max_length=255)),
                ('subdomain', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='AliasedTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Account')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Country',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='Manager',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Account')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Organization',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Project',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='account', to='tests.Account')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='ProjectManager',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Account')),
                ('manager', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Manager')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Project')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Record',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
                ('organization', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Organization')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SubTask',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('type', models.CharField(max_length=255)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Account')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Account')),
                ('project', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='tasks', to='tests.Project')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='UnscopedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
        ),
        migrations.CreateModel(
            name='SomeRelatedModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='TenantNotIdModel',
            fields=[
                ('tenant_column', models.IntegerField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AddField(
            model_name='somerelatedmodel',
            name='related_tenant',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.TenantNotIdModel'),
        ),

        migrations.AddField(
            model_name='subtask',
            name='task',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Task'),
        ),
        migrations.AddField(
            model_name='project',
            name='managers',
            field=models.ManyToManyField(through='tests.ProjectManager', to='tests.Manager'),
        ),
        migrations.AddField(
            model_name='aliasedtask',
            name='project_alias',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Project'),
        ),
        migrations.AddField(
            model_name='account',
            name='country',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Country'),
        ),
    ]



from django.db import connection
from django.test import TestCase, TransactionTestCase

from exam.cases import Exam
from exam.decorators import fixture, before, after

from .models import *


class Fixtures(Exam):
    # @after
    # def print_after(self):
    #     for q in connection.queries:
    #         print(q['sql'])

    @fixture
    def india(self):
        return Country.objects.create(name='India')

    @fixture
    def france(self):
        return Country.objects.create(name='France')

    @fixture
    def united_states(self):
        return Country.objects.create(name='United States')

    @fixture
    def account_fr(self):
        return Account.objects.create(pk=1,
                                      name='Account FR',
                                      country=self.france,
                                      subdomain='fr.',
                                      domain='citusdata.com')

    @fixture
    def account_in(self):
        return Account.objects.create(pk=2,
                                      name='Account IN',
                                      country=self.india,
                                      subdomain='in.',
                                      domain='citusdata.com')

    @fixture
    def account_us(self):
        return Account.objects.create(pk=3,
                                      name='Account US',
                                      country=self.united_states,
                                      subdomain='us.',
                                      domain='citusdata.com')


    @fixture
    def accounts(self):
        return [self.account_fr, self.account_in, self.account_us]


    @fixture
    def projects(self):
        projects = []

        for account in self.accounts:
            for i in range(10):
                projects.append(
                    Project.objects.create(account_id=account.pk,
                                           name='project %d' % i)
                )

        return projects


    @fixture
    def managers(self):
        managers = []

        for account in self.accounts:
            for i in range(5):
                managers.append(
                    Manager.objects.create(name='manager %d' % i,
                                           account=account)
                )

        return managers

    @fixture
    def tasks(self):
        tasks = []

        for project in self.projects:
            previous_task = None
            for i in range(5):
                previous_task = Task.objects.create(
                    name='task project %s %i' %(project.name, i),
                    project_id=project.pk,
                    account_id=project.account_id,
                    parent=previous_task)

                tasks.append(previous_task)

        return tasks

    @fixture
    def revenues(self):
        revenues = []

        for project in self.projects:
            for i in range(5):
                revenue = Revenue.objects.create(
                    value="%s mil" % i,
                    project_id=project.pk,
                    acc_id=project.account_id,
                )
                revenues.append(revenue)

        return revenues

    @fixture
    def project_managers(self):
        projects = self.projects
        managers = self.managers
        project_managers = []

        for project in projects:
            for manager in project.account.managers.all():
                project_managers.append(
                    ProjectManager.objects.create(account=project.account,
                                                  project=project,
                                                  manager=manager))
        return project_managers

    @fixture
    def subtasks(self):
        subtasks = []

        for task in self.tasks:
            for i in range(5):
                subtasks.append(
                    SubTask.objects.create(
                        name='subtask project %i, task %i',
                        type='test',
                        account_id=task.account_id,
                        project_id=task.project_id,
                        task=task)
                )

        return subtasks

    @fixture
    def unscoped(self):
        pass

    @fixture
    def aliased_tasks(self):
        pass

    @fixture
    def organization(self):
        return Organization.objects.create(name='organization')

    @fixture
    def tenant_not_id(self):
        tenants = []
        for i in range(3):
            tenant = TenantNotIdModel(tenant_column=i+1,
                                      name='test %d' % i)
            tenant.save()

            tenants.append(tenant)

            for j in range(10):
                SomeRelatedModel.objects.create(related_tenant=tenant,
                                                name='related %d' % j)
        return tenants


class BaseTestCase(Fixtures, TransactionTestCase):
    pass



# Generated by Django 2.2.9 on 2020-01-29 03:57

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0018_auto_20200128_0902'),
    ]

    atomic = False
    operations = [
        migrations.AddField(
            model_name='account',
            name='employee',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='accounts', to='tests.Employee'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2018-11-13 16:01
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings

from django_multitenant.db import migrations as tenant_migrations

def get_operations():
    operations = []
    if settings.USE_CITUS:
        operations = [
            # necessary for tests
            migrations.RunSQL("CREATE EXTENSION IF NOT EXISTS citus;"),
            migrations.RunSQL("SELECT * from master_add_node('django-multitenant_worker1_1', 5432);"),
            migrations.RunSQL("SELECT * from master_add_node('django-multitenant_worker2_1', 5432);")]


    operations += [
        # Drop constraints
        migrations.RunSQL("""
        ALTER TABLE tests_aliasedtask
        DROP CONSTRAINT tests_aliasedtask_pkey CASCADE;

        ALTER TABLE tests_aliasedtask ADD CONSTRAINT
        tests_aliasedtask_pkey PRIMARY KEY (account_id, id);
        """),
        migrations.RunSQL("ALTER TABLE tests_country DROP CONSTRAINT tests_country_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_manager DROP CONSTRAINT tests_manager_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_project DROP CONSTRAINT tests_project_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_projectmanager DROP CONSTRAINT tests_projectmanager_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_record DROP CONSTRAINT tests_record_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_subtask DROP CONSTRAINT tests_subtask_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_task DROP CONSTRAINT tests_task_pkey CASCADE;"),
    ]

    if settings.USE_CITUS:
        # distribute
        operations += [
            tenant_migrations.Distribute('Country', reference=True),
            tenant_migrations.Distribute('Account'),
            tenant_migrations.Distribute('AliasedTask'),
            tenant_migrations.Distribute('Manager'),
            tenant_migrations.Distribute('Organization'),
            tenant_migrations.Distribute('Project'),
            tenant_migrations.Distribute('ProjectManager'),
            tenant_migrations.Distribute('Record'),
            tenant_migrations.Distribute('SubTask'),
            tenant_migrations.Distribute('Task')]

    # Add constraints
    operations += [
        migrations.RunSQL("ALTER TABLE tests_country ADD CONSTRAINT tests_country_pkey PRIMARY KEY (id);"),
        migrations.RunSQL("ALTER TABLE tests_project ADD CONSTRAINT tests_project_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_manager ADD CONSTRAINT tests_manager_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_projectmanager ADD CONSTRAINT tests_projectmanager_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_record ADD CONSTRAINT tests_record_pkey PRIMARY KEY (organization_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_subtask ADD CONSTRAINT tests_subtask_pkey PRIMARY KEY (account_id, id);"),
        migrations.RunSQL("ALTER TABLE tests_task ADD CONSTRAINT tests_task_pkey PRIMARY KEY (account_id, id);")
    ]

    return operations


class Migration(migrations.Migration):
    atomic = False

    dependencies = [
        ('tests', '0001_initial'),
    ]

    operations = get_operations()



import re

from django.conf import settings
from django.db.utils import NotSupportedError, DataError

from django_multitenant.utils import (set_current_tenant,
                                      unset_current_tenant,
                                      get_current_tenant)

from .base import BaseTestCase


class TenantModelTest(BaseTestCase):
    def test_filter_without_joins(self):
        from .models import Project
        projects = self.projects
        account = self.account_fr

        self.assertEqual(Project.objects.count(), len(projects))
        set_current_tenant(account)
        self.assertEqual(Project.objects.count(), account.projects.count())
        unset_current_tenant()


    def test_filter_without_joins_on_tenant_id_not_pk(self):
        from .models import TenantNotIdModel, SomeRelatedModel
        tenants = self.tenant_not_id

        self.assertEqual(SomeRelatedModel.objects.count(), 30)
        set_current_tenant(tenants[0])
        self.assertEqual(SomeRelatedModel.objects.count(), 10)
        unset_current_tenant()

    def test_select_tenant(self):
        from .models import Project
        self.projects
        project = Project.objects.first()

        # Selecting the account, account_id being tenant
        with self.assertNumQueries(1) as captured_queries:
            account = project.account
            self.assertTrue('"tests_account"."id" = %d' % project.account_id \
                            in captured_queries.captured_queries[0]['sql'])


    def test_select_tenant_not_id(self):
        # Test with tenant_id not id field
        from .models import SomeRelatedModel

        tenants = self.tenant_not_id
        some_related = SomeRelatedModel.objects.first()

        # Selecting the tenant, tenant_column being tenant_id in tests_tenantnodidmodel
        with self.assertNumQueries(1) as captured_queries:
            tenant = some_related.related_tenant
            self.assertTrue('"tests_tenantnotidmodel"."tenant_column" = %d' % some_related.related_tenant_id \
                            in captured_queries.captured_queries[0]['sql'])

    def test_select_tenant_foreign_key(self):
        from .models import Task
        self.tasks

        task = Task.objects.first()
        set_current_tenant(task.account)

        # Selecting task.project, account is tenant
        # To push down, account_id should be in query
        with self.assertNumQueries(1) as captured_queries:
            project = task.project
            self.assertTrue('AND "tests_project"."account_id" = %d' % task.account_id \
                            in captured_queries.captured_queries[0]['sql'])


        unset_current_tenant()

    def test_select_tenant_foreign_key_different_tenant_id(self):
        from .models import Revenue, Account
        self.revenues

        revenue = Revenue.objects.first()
        set_current_tenant(revenue.acc)

        # Selecting revenue.project, project.account is tenant (revenue.acc is tenant)
        # To push down, account_id should be in query (not acc_id)
        with self.assertNumQueries(1) as captured_queries:
            project = revenue.project
            self.assertTrue('AND "tests_project"."account_id" = %d' % revenue.acc_id \
                            in captured_queries.captured_queries[0]['sql'])

        unset_current_tenant()

    def test_filter_select_related(self):
        from .models import Task
        task_id = self.tasks[0].pk

        # Test that the join clause has account_id
        with self.assertNumQueries(1) as captured_queries:
            task = Task.objects.filter(pk=task_id).select_related('project').first()

            self.assertEqual(task.account_id, task.project.account_id)
            pattern = 'INNER JOIN "tests_project" ON \("tests_task"."project_id" = "tests_project"."id" AND \("tests_task"."account_id" = .?"tests_project"."account_id"\)\)'
            self.assertTrue(bool(re.search(pattern, captured_queries.captured_queries[0]['sql'])))

    def test_filter_select_related_not_id_field(self):
        from .models import SomeRelatedModel
        tenants = self.tenant_not_id
        some_related_id = SomeRelatedModel.objects.first().pk
        some_related_tenant_id = SomeRelatedModel.objects.first().related_tenant_id

        # Test that the join clause has account_id
        with self.assertNumQueries(1) as captured_queries:
            related = SomeRelatedModel.objects.filter(pk=some_related_id).select_related('related_tenant').first()

            self.assertEqual(related.related_tenant_id, some_related_tenant_id)
            self.assertTrue('ON ("tests_somerelatedmodel"."related_tenant_id" = "tests_tenantnotidmodel"."tenant_column")' \
                            in captured_queries.captured_queries[0]['sql'])


    def test_prefetch_related(self):
        from .models import Project

        project_managers = self.project_managers
        project_id = project_managers[0].project_id
        account = project_managers[0].account
        set_current_tenant(account)

        with self.assertNumQueries(2) as captured_queries:
            project = Project.objects.filter(pk=project_id).prefetch_related('managers').first()

            self.assertTrue('WHERE ("tests_manager"."account_id" = %d' % project.account_id \
                            in captured_queries.captured_queries[1]['sql'])
            pattern = 'AND \("tests_projectmanager"."account_id" = .?"tests_manager"."account_id"\)'
            self.assertTrue(bool(re.search(pattern, captured_queries.captured_queries[1]['sql'])))

    def test_create_project(self):
        # Using save()
        from .models import Project
        account = self.account_fr

        project = Project()
        project.account = account
        project.name = 'test save()'

        project.save()

        self.assertEqual(Project.objects.count(), 1)

    def test_create_project_tenant_set(self):
        # Using save()
        from .models import Project
        account = self.account_fr

        set_current_tenant(account)
        project = Project()
        project.name = 'test save()'
        project.save()

        self.assertEqual(Project.objects.count(), 1)

        Project.objects.create(name='test create')
        self.assertEqual(Project.objects.count(), 2)

        unset_current_tenant()

    def test_bulk_create_tenant_set(self):
        from .models import Project
        account = self.account_fr

        set_current_tenant(account)
        projects = []
        for i in range(10):
            projects.append(Project(name='project %d' % i))

        Project.objects.bulk_create(projects)

        unset_current_tenant()
        self.assertEqual(Project.objects.count(), 10)
        for project in Project.objects.all():
            self.assertEqual(project.account_id, account.id)

    def test_bulk_create_tenant_not_set(self):
        from .models import Project
        account = self.account_fr

        unset_current_tenant()
        projects = []
        for i in range(10):
            projects.append(Project(name='project %d' % i,
                                    account=account))

        Project.objects.bulk_create(projects)

        self.assertEqual(Project.objects.count(), 10)
        for project in Project.objects.all():
            self.assertEqual(project.account_id, account.id)

        projects = []
        for i in range(10):
            projects.append(Project(name='project %d' % i))

        with self.assertRaises(DataError):
            Project.objects.bulk_create(projects)


    def test_update_tenant_project(self):
        if not settings.USE_CITUS:
            return

        from .models import Project
        account = self.account_fr

        project = Project()
        project.account = account
        project.name = 'test save()'

        project.save()

        self.assertEqual(Project.objects.count(), 1)

        project.account = self.account_in

        with self.assertRaises(NotSupportedError):
            project.save()

        project = Project.objects.first()
        self.assertEqual(project.account, account)

    def test_save_project(self):
        from .models import Project
        account = self.account_fr

        project = Project()
        project.account = account
        project.name = 'test save()'
        project.save()

        self.assertEqual(Project.objects.count(), 1)

        project.account = account
        project.name = 'test update'
        project.save()

        project = Project.objects.first()
        self.assertEqual(project.account, account)
        self.assertEqual(project.name, 'test update')

    def test_delete_tenant_set(self):
        from .models import Project
        projects = self.projects
        account = self.account_fr

        self.assertEqual(Project.objects.count(), 30)

        set_current_tenant(account)
        with self.assertNumQueries(7) as captured_queries:
            Project.objects.all().delete()
            unset_current_tenant()

            for query in captured_queries.captured_queries:
                if "tests_revenue" in query['sql']:
                    self.assertTrue('"acc_id" = %d' % account.id in query['sql'])
                else:
                    self.assertTrue('"account_id" = %d' % account.id in query['sql'])

        self.assertEqual(Project.objects.count(), 20)

    def test_delete_tenant_not_set(self):
        from .models import Project
        projects = self.projects

        self.assertEqual(Project.objects.count(), 30)
        Project.objects.all().delete()
        self.assertEqual(Project.objects.count(), 0)


    def test_subquery(self):
        # we want all the projects with the name of their first task
        import django
        if django.VERSION[0] == 1 and django.VERSION[1] < 11:
            # subqueries where only introduced in django 1.11
            return

        from django.db.models import OuterRef, Subquery
        from .models import Project, Task

        projects = self.projects
        account = self.account_fr
        tasks = self.tasks

        self.assertEqual(Project.objects.count(), 30)

        set_current_tenant(account)
        with self.assertNumQueries(1) as captured_queries:
            task_qs = Task.objects.filter(project=OuterRef("pk")).order_by('-name')
            projects = Project.objects.all().annotate(
                first_task_name=Subquery(
                    task_qs.values('name')[:1]
                )
            )
            for p in projects:
                self.assertTrue(p.first_task_name is not None)

            # check that tenant in subquery
            for query in captured_queries.captured_queries:
                self.assertTrue('U0."account_id" = %d' % account.id in query['sql'])
                self.assertTrue('WHERE "tests_project"."account_id" = %d' % account.id in query['sql'])

        unset_current_tenant()


    def test_subquery_joins(self):
        # we want all the projects with the name of their first task
        import django
        if django.VERSION[0] == 1 and django.VERSION[1] < 11:
            # subqueries where only introduced in django 1.11
            return

        from django.db.models import OuterRef, Subquery
        from .models import Project, SubTask

        projects = self.projects
        account = self.account_fr
        subtasks = self.subtasks

        self.assertEqual(Project.objects.count(), 30)

        set_current_tenant(account)
        with self.assertNumQueries(1) as captured_queries:
            subtask_qs = SubTask.objects.filter(project=OuterRef("pk"), task__opened=True).order_by('-name')
            projects = Project.objects.all().annotate(
                first_subtask_name=Subquery(
                    subtask_qs.values('name')[:1]
                )
            )
            for p in projects:
                self.assertTrue(p.first_subtask_name is not None)

            # check that tenant in subquery
            for query in captured_queries.captured_queries:
                self.assertTrue('U0."account_id" = %d' % account.id in query['sql'])

                pattern = '\(U0."task_id" = U\d."id" AND \(U0."account_id" = .?U\d."account_id"\)'
                self.assertTrue(bool(re.search(pattern, query['sql'])))
                self.assertTrue('WHERE "tests_project"."account_id" = %d' % account.id in query['sql'])

        unset_current_tenant()


    def test_task_manager(self):
        from .models import Task

        projects = self.projects
        account = self.account_fr
        tasks = self.tasks

        set_current_tenant(account)

        self.assertEqual(len(Task.objects.closed()), 0)
        self.assertEqual(len(Task.objects.opened()), 50)

        task = Task.objects.first()
        task.opened = False
        task.save()

        self.assertEqual(len(Task.objects.closed()), 1)
        self.assertEqual(len(Task.objects.opened()), 49)

        unset_current_tenant()


    def test_str_model_tenant_set(self):
        from .models import Task

        projects = self.projects
        account = self.account_fr
        tasks = self.tasks

        set_current_tenant(account)

        print(Task.objects.first())

        unset_current_tenant()


    def test_str_model_tenant_not_set(self):
        from .models import Task

        projects = self.projects
        account = self.account_fr
        tasks = self.tasks

        print(Task.objects.first())

    def test_exclude_tenant_set(self):
        from .models import Task
        projects = self.projects
        account_fr = self.account_fr
        tasks = self.tasks

        unset_current_tenant()
        set_current_tenant(account_fr)

        tasks = Task.objects.exclude(project__isnull=True)

        self.assertEqual(tasks.count(), 50)

        unset_current_tenant()

    def test_exclude_tenant_not_set(self):
        from .models import Task
        projects = self.projects
        account_fr = self.account_fr
        tasks = self.tasks

        tasks = Task.objects.exclude(project__isnull=True)
        self.assertEqual(tasks.count(), 150)

    def test_exclude_related(self):
        from .models import Project, Manager, ProjectManager
        project = self.projects[0]
        project_managers = self.project_managers
        account = project.account
        manager = Manager.objects.create(name='Louise', account=account)
        ProjectManager.objects.create(account=account, project=project, manager=manager)

        excluded = Project.objects.exclude(projectmanagers__manager__name='Louise')
        self.assertEqual(excluded.count(), 29)

    def test_delete_cascade_distributed(self):
        from .models import Task, Project, SubTask
        subtasks = self.subtasks
        account = self.account_fr

        unset_current_tenant()

        self.assertEqual(Project.objects.count(), 30)
        self.assertEqual(Task.objects.count(), 150)
        self.assertEqual(SubTask.objects.count(), 750)

        set_current_tenant(account)

        self.assertEqual(Project.objects.count(), 10)
        self.assertEqual(Task.objects.count(), 50)
        self.assertEqual(SubTask.objects.count(), 250)

        project = Project.objects.first()

        with self.assertNumQueries(12) as captured_queries:
            project.delete()

            self.assertEqual(Project.objects.count(), 9)
            self.assertEqual(Task.objects.count(), 45)
            self.assertEqual(SubTask.objects.count(), 225)
            self.assertFalse("SET LOCAL citus.multi_shard_modify_mode TO 'sequential';" in [query['sql'] for query in captured_queries.captured_queries])
            self.assertFalse("SET LOCAL citus.multi_shard_modify_mode TO 'parallel';" in [query['sql'] for query in captured_queries.captured_queries])

    def test_delete_cascade_reference_to_distributed(self):
        from .models import Country, Account
        unset_current_tenant()

        country = self.france
        account1 = Account.objects.create(name='Account FR',
                                          country=country,
                                          subdomain='fr.',
                                          domain='citusdata.com')

        account2 = Account.objects.create(name='Account FR 2',
                                          country=country,
                                          subdomain='fr.',
                                          domain='msft.com')

        self.assertEqual(Account.objects.count(), 2)

        with self.assertNumQueries(16) as captured_queries:
            country.delete()

            self.assertEqual(Account.objects.count(), 0)
            self.assertEqual(Country.objects.count(), 0)
            self.assertTrue("SET LOCAL citus.multi_shard_modify_mode TO 'sequential';" in [query['sql'] for query in captured_queries.captured_queries])
            self.assertTrue("SET LOCAL citus.multi_shard_modify_mode TO 'parallel';" in [query['sql'] for query in captured_queries.captured_queries])

    def test_delete_cascade_distributed_to_reference(self):
        from .models import Account, Employee, ModelConfig, Project
        unset_current_tenant()

        account = self.account_fr
        employee = Employee.objects.create(account=account, name='Louise')
        modelconfig = ModelConfig.objects.create(account=account,
                                                 employee=employee,
                                                 name='test')
        projects = self.projects


        for project in projects:
            if project.account == account:
                project.employee = employee
                project.save(update_fields=['employee'])

        account.employee = employee
        account.save()

        self.assertEqual(Account.objects.count(), 3)
        self.assertEqual(Employee.objects.count(), 1)
        self.assertEqual(ModelConfig.objects.count(), 1)
        self.assertEqual(Project.objects.count(), 30)

        set_current_tenant(account)
        with self.assertNumQueries(28) as captured_queries:
            account.delete()

            # Once deleted, we don't have a current tenant
            self.assertEqual(Account.objects.count(), 2)
            self.assertEqual(Employee.objects.count(), 0)
            self.assertEqual(ModelConfig.objects.count(), 0)
            self.assertEqual(Project.objects.count(), 20)
            self.assertTrue("SET LOCAL citus.multi_shard_modify_mode TO 'sequential';" in [query['sql'] for query in captured_queries.captured_queries])
            self.assertTrue("SET LOCAL citus.multi_shard_modify_mode TO 'parallel';" in [query['sql'] for query in captured_queries.captured_queries])

        unset_current_tenant()


class MultipleTenantModelTest(BaseTestCase):
    def test_filter_without_joins(self):
        from .models import Project, Account
        unset_current_tenant()
        projects = self.projects
        accounts = Account.objects.all().order_by('id')[1:]

        self.assertEqual(Project.objects.count(), 30)
        set_current_tenant(accounts)
        self.assertEqual(Project.objects.count(), 20)

        unset_current_tenant()

    def test_filter_without_joins_on_tenant_id_not_pk(self):
        from .models import TenantNotIdModel, SomeRelatedModel
        unset_current_tenant()
        tenants = self.tenant_not_id

        self.assertEqual(SomeRelatedModel.objects.count(), 30)
        set_current_tenant([tenants[0], tenants[1]])
        self.assertEqual(SomeRelatedModel.objects.count(), 20)

        unset_current_tenant()

    def test_select_tenant_foreign_key(self):
        from .models import Task, Account
        unset_current_tenant()
        self.tasks

        task = Task.objects.first()
        accounts = [task.account, Account.objects.last()]
        set_current_tenant(accounts)

        # Selecting task.project, account is tenant
        # To push down, account_id should be in query
        with self.assertNumQueries(1) as captured_queries:
            project = task.project
            self.assertTrue('AND "tests_project"."account_id" IN (%s)' % ', '.join([str(account.id) for account in accounts]) \
                            in captured_queries.captured_queries[0]['sql'])

        unset_current_tenant()

    def test_select_tenant_foreign_key_different_tenant_id(self):
        from .models import Revenue, Account
        unset_current_tenant()
        self.revenues

        revenue = Revenue.objects.first()
        accounts = [revenue.acc, Account.objects.last()]
        set_current_tenant(accounts)

        # Selecting revenue.project, project.account is tenant (revenue.acc is tenant)
        # To push down, account_id should be in query (not acc_id)
        with self.assertNumQueries(1) as captured_queries:
            project = revenue.project
            self.assertTrue('AND "tests_project"."account_id" IN (%s)' % ', '.join([str(account.id) for account in accounts]) \
                            in captured_queries.captured_queries[0]['sql'])

        unset_current_tenant()

    def test_prefetch_related(self):
        from .models import Project, Account
        unset_current_tenant()
        project_managers = self.project_managers
        project_id = project_managers[0].project_id
        accounts = [project_managers[0].account, Account.objects.last()]
        set_current_tenant(accounts)

        with self.assertNumQueries(2) as captured_queries:
            project = Project.objects.filter(pk=project_id).prefetch_related('managers').first()

            self.assertTrue('WHERE ("tests_manager"."account_id" IN (%s)' % ', '.join([str(account.id) for account in accounts]) \
                            in captured_queries.captured_queries[1]['sql'])

            pattern = 'AND \("tests_projectmanager"."account_id" = .?"tests_manager"."account_id"\)'
            self.assertTrue(bool(re.search(pattern, captured_queries.captured_queries[1]['sql'])))

        unset_current_tenant()

    def test_delete_tenant_set(self):
        from .models import Project, Account
        unset_current_tenant()
        projects = self.projects
        accounts = Account.objects.all().order_by('id')[1:]
        self.assertEqual(Project.objects.count(), 30)

        set_current_tenant(accounts)
        with self.assertNumQueries(8) as captured_queries:
            Project.objects.all().delete()

            for query in captured_queries.captured_queries:
                if "tests_revenue" in query['sql']:
                    self.assertTrue('"acc_id" IN (%s)' % ', '.join([str(account.id) for account in accounts]))
                else:
                    self.assertTrue('"account_id" IN (%s)' % ', '.join([str(account.id) for account in accounts]))

        unset_current_tenant()
        self.assertEqual(Project.objects.count(), 10)


    def test_subquery(self):
        # subquery don't work for multi tenants
        # we want all the projects with the name of their first task
        pass


    def test_uuid_create(self):
        from .models import Record
        organization = self.organization
        set_current_tenant(organization)
        record = Record.objects.create(name='record1')
        self.assertEqual(record.organization_id, organization.id)

    def test_uuid_save(self):
        from .models import Record
        organization = self.organization
        set_current_tenant(organization)
        record = Record(name='record1')
        record.save()

        record = Record.objects.first()
        self.assertEqual(record.organization_id, organization.id)

    def test_save_tenant_unset(self):
        unset_current_tenant()
        from .models import Project
        account = self.account_fr

        account_2 = self.account_us

        project = Project(account=account, name='test save fr')
        project.save()

        project2 = Project(account=account_2, name='test save us')
        project2.save()

        self.assertEqual(Project.objects.count(), 2)

        project.name = 'test update name'
        project.save()

        project = Project.objects.first()
        self.assertEqual(project.name, 'test update name')

    def test_save_tenant_set_different_than_object(self):
        unset_current_tenant()
        from .models import Project
        account = self.account_fr
        account_2 = self.account_us

        project = Project(account=account, name='test save fr')
        project.save()

        project2 = Project(account=account_2, name='test save us')
        project2.save()

        self.assertEqual(Project.objects.count(), 2)

        set_current_tenant(account_2)

        project.name = 'test update name'
        project.save()

        current_tenant = get_current_tenant()
        self.assertEqual(current_tenant, account_2)

        unset_current_tenant()
        project = Project.objects.filter(account=account).first()
        self.assertEqual(project.name, 'test update name')

    def test_save_tenant_set(self):
        unset_current_tenant()
        from .models import Project
        account = self.account_fr
        account_2 = self.account_us

        project = Project(account=account, name='test save fr')
        project.save()

        project2 = Project(account=account_2, name='test save us')
        project2.save()

        self.assertEqual(Project.objects.count(), 2)

        set_current_tenant(account)

        project.name = 'test update name'
        project.save()

        current_tenant = get_current_tenant()
        self.assertEqual(current_tenant, account)

        unset_current_tenant()
        project = Project.objects.filter(account=account).first()
        self.assertEqual(project.name, 'test update name')



import logging

from django.db import models
from django.db.models.sql import DeleteQuery, UpdateQuery
from django.db.models.deletion import Collector
from django.db.utils import NotSupportedError

from .deletion import related_objects
from .query import wrap_get_compiler, wrap_update_batch, wrap_delete
from .utils import (
    set_current_tenant,
    get_current_tenant,
    get_current_tenant_value,
    get_tenant_field,
    get_tenant_filters,
    get_object_tenant,
    set_object_tenant
)


logger = logging.getLogger(__name__)


class TenantManagerMixin(object):
    #Below is the manager related to the above class.
    def get_queryset(self):
        #Injecting tenant_id filters in the get_queryset.
        #Injects tenant_id filter on the current model for all the non-join/join queries. 
        queryset = self._queryset_class(self.model)
        current_tenant = get_current_tenant()
        if current_tenant:
            kwargs = get_tenant_filters(self.model)
            return queryset.filter(**kwargs)
        return queryset

    def bulk_create(self, objs, **kwargs):
        if get_current_tenant():
            tenant_value = get_current_tenant_value()
            for obj in objs:
                set_object_tenant(obj, tenant_value)

        return super(TenantManagerMixin, self).bulk_create(objs, **kwargs)


class TenantModelMixin(object):
    #Abstract model which all the models related to tenant inherit.
    tenant_id = ''

    def __init__(self, *args, **kwargs):
        if not hasattr(DeleteQuery.get_compiler, "_sign"):
            DeleteQuery.get_compiler = wrap_get_compiler(DeleteQuery.get_compiler)
            Collector.related_objects = related_objects
            Collector.delete = wrap_delete(Collector.delete)

        if not hasattr(UpdateQuery.get_compiler, "_sign"):
            UpdateQuery.update_batch = wrap_update_batch(UpdateQuery.update_batch)

        super(TenantModelMixin, self).__init__(*args, **kwargs)

    def __setattr__(self, attrname, val):

        if (attrname in (self.tenant_field, get_tenant_field(self).name)
            and not self._state.adding
            and val
            and self.tenant_value
            and val != self.tenant_value
            and val != self.tenant_object):
            self._try_update_tenant = True

        return super(TenantModelMixin, self).__setattr__(attrname, val)

    def _do_update(self, base_qs, using, pk_val, values, update_fields, forced_update):
        #adding tenant filters for save
        #Citus requires tenant_id filters for update, hence doing this below change.

        current_tenant = get_current_tenant()

        if current_tenant:
            kwargs = get_tenant_filters(self.__class__)
            base_qs = base_qs.filter(**kwargs)
        else:
            logger.warning('Attempting to update %s instance "%s" without a current tenant '
                           'set. This may cause issues in a partitioned environment. '
                           'Recommend calling set_current_tenant() before performing this '
                           'operation.',
                           self._meta.model.__name__, self)

        return super(TenantModelMixin,self)._do_update(base_qs, using,
                                                       pk_val, values,
                                                       update_fields,
                                                       forced_update)

    def save(self, *args, **kwargs):
        if hasattr(self, '_try_update_tenant'):
            raise NotSupportedError('Tenant column of a row cannot be updated.')

        current_tenant = get_current_tenant()
        tenant_value = get_current_tenant_value()

        set_object_tenant(self, tenant_value)

        if self.tenant_value and tenant_value != self.tenant_value:
            self_tenant = get_object_tenant(self)
            set_current_tenant(self_tenant)

        obj = super(TenantModelMixin, self).save(*args, **kwargs)

        set_current_tenant(current_tenant)

        return obj

    @property
    def tenant_field(self):
        return self.tenant_id

    @property
    def tenant_value(self):
        return getattr(self, self.tenant_field, None)

    @property
    def tenant_object(self):
        return get_object_tenant(self)



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-04-12 08:39
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0008_auto_20190412_0831'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='somerelatedmodel',
            name='opened',
        ),
        migrations.AddField(
            model_name='task',
            name='opened',
            field=models.BooleanField(default=True),
        ),
    ]



# Generated by Django 2.2.9 on 2020-01-28 09:02

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields
import django_multitenant.mixins

from django_multitenant.db import migrations as tenant_migrations


def get_operations():
    operations = [
        migrations.CreateModel(
            name='ModelConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('account', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='configs', to='tests.Account')),
                ('employee', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='configs', to='tests.Employee')),
            ],
            options={
                'abstract': False,
            },
            bases=(django_multitenant.mixins.TenantModelMixin, models.Model),
        ),

        migrations.RunSQL("ALTER TABLE tests_modelconfig DROP CONSTRAINT tests_modelconfig_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_modelconfig ADD CONSTRAINT tests_modelconfig_pkey PRIMARY KEY (account_id, id);"),
    ]

    if settings.USE_CITUS:
        operations += [tenant_migrations.Distribute('ModelConfig'),]

    return operations


class Migration(migrations.Migration):

    atomic = False

    dependencies = [
        ('tests', '0017_auto_20200128_0853'),
    ]

    operations = get_operations()



from django.db import connection
from django.db.migrations.migration import Migration
from django.db.migrations.recorder import MigrationRecorder
from django.db.migrations.state import ModelState, ProjectState
from django.conf import settings
from django.test import TransactionTestCase

from django_multitenant.db import migrations



if settings.USE_CITUS:
    class MigrationTest(TransactionTestCase):
        def undistribute_table(self, table_name):
            queries = [
                "CREATE TABLE %(table_name)s_bis (LIKE %(table_name)s INCLUDING ALL);"
                "CREATE TEMP TABLE %(table_name)s_temp AS SELECT * FROM %(table_name)s;"
                "INSERT INTO %(table_name)s_bis SELECT * FROM %(table_name)s_temp;"
                "DROP TABLE %(table_name)s CASCADE;"
                "ALTER TABLE %(table_name)s_bis RENAME TO %(table_name)s;"]

            with connection.cursor() as cursor:
                for query in queries:
                    cursor.execute(query % {'table_name': table_name})


        def assertTableIsDistributed(self, table_name, column_name, value=True):
            query = """
            SELECT logicalrelid, pg_attribute.attname
            FROM pg_dist_partition
            INNER JOIN pg_attribute ON (logicalrelid=attrelid)
            WHERE logicalrelid::varchar(255) = '{}'
            AND partmethod='h'
            AND attnum=substring(partkey from '%:varattno #"[0-9]+#"%' for '#')::int;
            """
            distributed = False
            with connection.cursor() as cursor:
                cursor.execute(query.format(table_name))
                row = cursor.fetchone()

                if row and row[1] == column_name:
                    distributed = True

                self.assertEqual(distributed, value)

        def assertTableIsNotDistributed(self, table_name, column_name, value=True):
            return self.assertTableIsDistributed(table_name, column_name, value=False)

        def assertTableIsReference(self, table_name, value=False):
            query = """
            SELECT 1
            FROM pg_dist_partition
            WHERE logicalrelid::varchar(255) = %s
            AND partmethod='m';
            """
            distributed = False

            with connection.cursor() as cursor:
                cursor.execute(query, [table_name])
                row = cursor.fetchone()

                distributed = True if row else False

            self.assertEqual(distributed, value)

        def assertTableIsNotReference(self, table_name, value=False):
            return self.assertTableIsReference(table_name, value=False)

        def test_distribute_table(self):
            project_state = ProjectState(real_apps=['tests'])
            operation = migrations.Distribute('MigrationTestModel')

            self.assertEqual(operation.describe(), "Run create_(distributed/reference)_table statement")

            self.assertTableIsNotDistributed('tests_migrationtestmodel', 'id')
            new_state = project_state.clone()

            with connection.schema_editor() as editor:
                operation.database_forwards("tests", editor, project_state, new_state)

            self.assertTableIsDistributed('tests_migrationtestmodel', 'id')
            self.undistribute_table('tests_migrationtestmodel')

        def test_reference_table(self):
            project_state = ProjectState(real_apps=['tests'])
            operation = migrations.Distribute('MigrationTestReferenceModel', reference=True)

            self.assertEqual(operation.describe(), "Run create_(distributed/reference)_table statement")
            self.assertTableIsNotReference('tests_migrationtestreferencemodel')
            new_state = project_state.clone()

            with connection.schema_editor() as editor:
                operation.database_forwards("tests", editor, project_state, new_state)

            self.assertTableIsReference('tests_migrationtestreferencemodel')
            self.undistribute_table('tests_migrationtestreferencemodel')

        def test_reference_different_app_table(self):
            project_state = ProjectState(real_apps=['auth'])
            operation = migrations.Distribute('auth.User', reference=True)

            self.assertEqual(operation.describe(), "Run create_(distributed/reference)_table statement")
            self.assertTableIsNotReference('auth_user')
            new_state = project_state.clone()

            with connection.schema_editor() as editor:
                operation.database_forwards("tests", editor, project_state, new_state)

            self.assertTableIsReference('auth_user')
            self.undistribute_table('auth_user')



# -*- coding: utf-8 -*-
# Generated by Django 1.11.23 on 2019-10-25 08:44
from __future__ import unicode_literals

from django.db import migrations
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0015_auto_20190829_1334'),
    ]

    operations = [
        migrations.AlterField(
            model_name='projectmanager',
            name='project',
            field=django_multitenant.fields.TenantForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='projectmanagers', to='tests.Project'),
        ),
    ]



from django.db.backends.utils import CursorWrapper as BaseCursorWrapper


class CursorWrapper(BaseCursorWrapper):
    def _execute(self, sql, params, *ignored_wrapper_args):
        self.db.validate_no_broken_transaction()
        with self.db.wrap_database_errors:
            if params is None:
                try:
                    return self.cursor.execute(sql)
                except:
                    import ipdb; ipdb.set_trace()
                    print(sql)
            else:
                try:
                    return self.cursor.execute(sql, params)
                except:
                    import ipdb; ipdb.set_trace()
                    print(sql)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.22 on 2019-08-28 13:14
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields
import django_multitenant.mixins


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0013_auto_20190517_1607'),
    ]

    operations = [
        migrations.CreateModel(
            name='MigrationTestModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
            ],
            options={
                'abstract': False,
            },
            bases=(django_multitenant.mixins.TenantModelMixin, models.Model),
        ),
    ]



from .settings import *

DATABASES = {
    "default": {
        'ENGINE': 'django_multitenant.backends.postgresql',
        "NAME": "postgres",
        "USER": "postgres",
        "PASSWORD": "",
        "HOST": "localhost",
        "PORT": 5604,
        "TEST": {
            "NAME": "postgres",
            "SERIALIZE": False
        }
    }
}


USE_CITUS = False



# -*- coding: utf-8 -*-
import os
import django


BASE_PATH = os.path.normpath(
    os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir)
)

DATABASES = {
    "default": {
        'ENGINE': 'django_multitenant.backends.postgresql',
        "NAME": "postgres",
        "USER": "postgres",
        "PASSWORD": "",
        "HOST": "localhost",
        "PORT": 5600,
        "TEST": {
            "NAME": "postgres",
            "SERIALIZE": False
        }
    }
}

SITE_ID = 1
DEBUG = True

MIDDLEWARE_CLASSES = (
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.locale.LocaleMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)


MIDDLEWARE = MIDDLEWARE_CLASSES

INSTALLED_APPS = [
    "django.contrib.admin",
    "django.contrib.messages",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django_multitenant",
    "django_multitenant.tests",
]

SECRET_KEY = "blabla"

ROOT_URLCONF = "django_multitenant.tests.urls"


TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
                'django.template.context_processors.i18n',
            ],
        },
    },
]

USE_CITUS = True
CITUS_EXTENSION_INSTALLED = True



# Generated by Django 2.1.1 on 2019-03-01 08:34

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0003_auto_20181220_1151'),
    ]

    operations = [
        migrations.AlterField(
            model_name='manager',
            name='account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='managers', to='tests.Account'),
        ),
        migrations.AlterField(
            model_name='project',
            name='account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='projects', to='tests.Account'),
        ),
    ]



# Generated by Django 2.2.9 on 2020-01-28 08:53

from django.db import migrations, models
import django.db.models.deletion
from django.conf import settings
import django_multitenant.fields

from django_multitenant.db import migrations as tenant_migrations


def get_operations():
    operations = [
        migrations.CreateModel(
            name='Employee',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('account', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='employees', db_constraint=False, to='tests.Account')),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='users_created', to='tests.Employee')),
            ],
        ),
    ]

    if settings.USE_CITUS:
        operations += [tenant_migrations.Distribute('Employee', reference=True),]

    return operations

class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0016_auto_20191025_0844'),
    ]

    operations = get_operations()



# Generated by Django 2.2.9 on 2020-01-29 04:53

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0020_auto_20200129_0404'),
    ]
    atomic = False
    operations = [
        migrations.AlterField(
            model_name='modelconfig',
            name='employee',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='configs', to='tests.Employee'),
        )
    ]



from django_multitenant.utils import (set_current_tenant,
                                      get_current_tenant,
                                      unset_current_tenant,
                                      get_tenant_column,
                                      get_current_tenant_value,
                                      get_tenant_filters)

from .base import BaseTestCase


class UtilsTest(BaseTestCase):
    def test_set_current_tenant(self):
        from .models import Project
        projects = self.projects
        account = projects[0].account

        set_current_tenant(account)
        self.assertEqual(get_current_tenant(), account)
        unset_current_tenant()

    def test_get_tenant_column(self):
        from .models import Project
        projects = self.projects
        account = projects[0].account

        # test if instance
        column = get_tenant_column(account)
        self.assertEqual(column, 'id')
        self.assertEqual(get_tenant_column(account), account.tenant_field)

        # test if  model
        column = get_tenant_column(Project)
        self.assertEqual(column, 'account_id')

    def test_current_tenant_value_single(self):
        from .models import Project, Account

        projects = self.projects
        account = projects[0].account
        set_current_tenant(account)

        self.assertEqual(get_current_tenant_value(), account.id)

        unset_current_tenant()

    def test_current_tenant_value_list(self):
        from .models import Project, Account

        projects = self.projects
        accounts = [projects[0].account, projects[1].account]
        set_current_tenant(accounts)

        value = get_current_tenant_value()

        self.assertTrue(isinstance(value, list))
        self.assertEqual(value, [accounts[0].id, accounts[1].id])

        unset_current_tenant()

    def test_current_tenant_value_queryset(self):
        from .models import Project, Account

        projects = self.projects
        accounts = Account.objects.all().order_by('id')
        set_current_tenant(accounts)

        value = get_current_tenant_value()

        self.assertTrue(isinstance(value, list))
        self.assertEqual(value, accounts.values_list('id', flat=True))

        unset_current_tenant()

    def test_tenant_filters_single_tenant(self):
        from .models import Project, Account

        projects = self.projects
        account = projects[0].account
        set_current_tenant(account)

        self.assertEqual(get_tenant_filters(Project), {'account_id': account.pk})

        unset_current_tenant()

    def test_tenant_filters_multi_tenant(self):
        from .models import Project, Account

        projects = self.projects
        accounts = [projects[0].account, projects[1].account]
        set_current_tenant(accounts)

        self.assertEqual(get_tenant_filters(Project),
                         {'account_id__in': [accounts[0].id, accounts[1].id]})

        unset_current_tenant()

    def test_current_tenant_value_queryset(self):
        from .models import Project, Account

        projects = self.projects
        accounts = Account.objects.all().order_by('id')
        set_current_tenant(accounts)

        value = get_current_tenant_value()

        self.assertEqual(get_tenant_filters(Project),
                         {'account_id__in': list(accounts.values_list('id', flat=True))})
        unset_current_tenant()



import uuid

from django.db import models
from django.db.models.signals import pre_save
from django.dispatch import receiver


from django_multitenant.mixins import TenantModelMixin, TenantManagerMixin
from django_multitenant.models import TenantModel
from django_multitenant.fields import TenantForeignKey


class Country(models.Model):
    name = models.CharField(max_length=255)


class Account(TenantModel):
    name = models.CharField(max_length=255)
    domain = models.CharField(max_length=255)
    subdomain = models.CharField(max_length=255)
    country = models.ForeignKey(Country, on_delete=models.CASCADE)
    employee = models.ForeignKey('Employee', on_delete=models.CASCADE,
                                 related_name='accounts',
                                 null=True,
                                 blank=True)


    # TODO change to Meta
    tenant_id = 'id'

    def __str__(self):
        return "{}".format(self.name)


class Employee(models.Model):
    # Reference table
    account = models.ForeignKey(Account,
                                on_delete=models.CASCADE,
                                null=True,
                                blank=True,
                                related_name='employees')
    created_by = models.ForeignKey('self',
                                   blank=True,
                                   null=True,
                                   related_name='users_created',
                                   on_delete=models.SET_NULL)

    name = models.CharField(max_length=255)


class ModelConfig(TenantModel):
    name = models.CharField(max_length=255)
    account = models.ForeignKey(Account, on_delete=models.CASCADE,
                                related_name='configs')
    employee = models.ForeignKey(Employee, on_delete=models.CASCADE,
                                 related_name='configs',
                                 null=True,
                                 blank=True)

    tenant_id = 'account_id'


class Manager(TenantModel):
    name = models.CharField(max_length=255)
    account = models.ForeignKey(Account, on_delete=models.CASCADE,
                                related_name='managers')
    tenant_id = 'account_id'


class Project(TenantModel):
    name = models.CharField(max_length=255)
    account = models.ForeignKey(Account, related_name='projects',
                                on_delete=models.CASCADE)
    managers = models.ManyToManyField(Manager, through='ProjectManager')
    employee = models.ForeignKey(Employee, related_name='projects',
                                 on_delete=models.SET_NULL,
                                 null=True,
                                 blank=True)
    tenant_id = 'account_id'

    def __str__(self):
        return "{} ({})".format(self.name, self.account)


class ProjectManager(TenantModel):
    project = TenantForeignKey(Project, on_delete=models.CASCADE,
                               related_name='projectmanagers')
    manager = TenantForeignKey(Manager, on_delete=models.CASCADE)
    account = models.ForeignKey(Account, on_delete=models.CASCADE)

    tenant_id = 'account_id'


class TaskQueryset(models.QuerySet):
    def opened(self):
        return self.filter(opened=True)

    def closed(self):
        return self.filter(opened=False)


class TaskManager(TenantManagerMixin, models.Manager):
    _queryset_class = TaskQueryset

    def opened(self):
        return self.get_queryset().opened()

    def closed(self):
        return self.get_queryset().closed()


class Task(TenantModelMixin, models.Model):
    name = models.CharField(max_length=255)
    project = TenantForeignKey(Project, on_delete=models.CASCADE,
                               related_name='tasks')
    account = models.ForeignKey(Account, on_delete=models.CASCADE)
    opened = models.BooleanField(default=True)

    parent = TenantForeignKey('self', on_delete=models.CASCADE, db_index=False,
                              blank=True, null=True)

    objects = TaskManager()

    tenant_id = 'account_id'


    def __str__(self):
        return "{} ({})".format(self.name, self.project)


class SubTask(TenantModel):
    name = models.CharField(max_length=255)
    type = models.CharField(max_length=255)
    account = models.ForeignKey(Account, on_delete=models.CASCADE)
    task = TenantForeignKey(Task, on_delete=models.CASCADE)
    project = TenantForeignKey(Project, on_delete=models.CASCADE, null=True)

    tenant_id = 'account_id'

class UnscopedModel(models.Model):
    name = models.CharField(max_length=255)


class AliasedTask(TenantModel):
    project_alias = TenantForeignKey(Project, on_delete=models.CASCADE)
    account = models.ForeignKey(Account, on_delete=models.CASCADE)

    tenant_id = 'account_id'


class Revenue(TenantModel):
    # To test for correct tenant_id push down in query
    acc = models.ForeignKey(Account, on_delete=models.CASCADE,
                            related_name='revenues')
    project = TenantForeignKey(Project, on_delete=models.CASCADE,
                               related_name='revenues')
    value = models.CharField(max_length=30)

    tenant_id = 'acc_id'


# Models for UUID tests
class Organization(TenantModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    tenant_id = 'id'


class Record(TenantModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=255)
    organization = TenantForeignKey(Organization, on_delete=models.CASCADE)

    tenant_id = 'organization_id'



class TenantNotIdModel(TenantModel):
    tenant_column = models.IntegerField(primary_key=True, editable=False)
    name = models.CharField(max_length=255)

    tenant_id = 'tenant_column'


class SomeRelatedModel(TenantModel):
    related_tenant = models.ForeignKey(TenantNotIdModel, on_delete=models.CASCADE)
    name = models.CharField(max_length=255)

    tenant_id = 'related_tenant_id'


class MigrationTestModel(TenantModel):
    name = models.CharField(max_length=255)
    tenant_id = 'id'


class MigrationTestReferenceModel(models.Model):
    name = models.CharField(max_length=255)



import logging
from django.apps import apps
from django.db.backends.postgresql.base import (
    DatabaseWrapper as PostgresqlDatabaseWrapper,
    DatabaseSchemaEditor as PostgresqlDatabaseSchemaEditor,
    DatabaseCreation,
    DatabaseFeatures,
    DatabaseOperations,
    DatabaseClient,
    DatabaseIntrospection
)
from django_multitenant.fields import TenantForeignKey
from django_multitenant.utils import get_model_by_db_table, get_tenant_column

logger = logging.getLogger(__name__)


class DatabaseSchemaEditor(PostgresqlDatabaseSchemaEditor):
    # Override
    def __enter__(self):
        ret = super(DatabaseSchemaEditor, self).__enter__()
        return ret

    # Override
    def _alter_field(self, model, old_field, new_field, old_type, new_type,
                     old_db_params, new_db_params, strict=False):

        super(DatabaseSchemaEditor, self)._alter_field(model, old_field,
                                                       new_field, old_type,
                                                       new_type, old_db_params,
                                                       new_db_params, strict)

        # If the pkey was dropped in the previous distribute migration,
        # foreign key constraints didn't previously exists so django does not
        # recreated them.
        # Here we test if we are in this case
        if isinstance(new_field, TenantForeignKey) and new_field.db_constraint:
            from_model = get_model_by_db_table(model._meta.db_table)
            fk_names = (self._constraint_names(model,
                                               [new_field.column],
                                               foreign_key=True) +
                        self._constraint_names(model,
                                               [new_field.column,
                                                get_tenant_column(from_model)],
                                               foreign_key=True))
            if not fk_names:
                self.execute(self._create_fk_sql(model, new_field,
                                                 "_fk_%(to_table)s_%(to_column)s"))

    # Override
    def _create_fk_sql(self, model, field, suffix):
        if isinstance(field, TenantForeignKey):
            try:
                # test if both models exists
                # This case happens when we are running from scratch migrations and one model was removed from code
                # In the previous migrations we would still be creating the foreign key
                from_model = get_model_by_db_table(model._meta.db_table)
                to_model = get_model_by_db_table(field.target_field.model._meta.db_table)
            except ValueError:
                return None

            from_columns = field.column, get_tenant_column(from_model)
            to_columns = field.target_field.column, get_tenant_column(to_model)
            suffix = suffix % {
                "to_table": field.target_field.model._meta.db_table,
                "to_column": '_'.join(to_columns),
            }

            return self.sql_create_fk % {
                "table": self.quote_name(model._meta.db_table),
                "name": self.quote_name(self._create_index_name(model, from_columns, suffix=suffix)),
                "column": ', '.join([self.quote_name(from_col) for from_col in from_columns]),
                "to_table": self.quote_name(field.target_field.model._meta.db_table),
                "to_column": ', '.join([self.quote_name(to_col) for to_col in to_columns]),
                "deferrable": self.connection.ops.deferrable_sql(),
            }
        return super(DatabaseSchemaEditor, self)._create_fk_sql(model, field, suffix)

    # Override
    def execute(self, sql, params=()):
        # Hack: Citus will throw the following error if these statements are
        # not executed separately: "ERROR: cannot execute multiple utility events"
        if sql and not params:
            for statement in str(sql).split(';'):
                if statement and not statement.isspace():
                    super(DatabaseSchemaEditor, self).execute(statement)
        elif sql:
            super(DatabaseSchemaEditor, self).execute(sql, params)


    def _create_index_name(self, model, column_names, suffix=""):
        # compat with django 2.X and django 1.X
        import django

        if not isinstance(model, str) and django.VERSION[0] > 1:
            model = model._meta.db_table

        return super(DatabaseSchemaEditor, self)._create_index_name(model,
                                                                    column_names,
                                                                    suffix=suffix)


class DatabaseWrapper(PostgresqlDatabaseWrapper):
    # Override
    SchemaEditorClass = DatabaseSchemaEditor



# Generated by Django 2.2.9 on 2020-01-29 04:04

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0019_auto_20200129_0357'),
    ]

    atomic = False

    operations = [
        migrations.AddField(
            model_name='project',
            name='employee',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='projects', to='tests.Employee'),
        ),
    ]



from setuptools import setup, find_packages
from os import path
from io import open


# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name="django-multitenant",
    version='2.3.1',  # Required
    description='Django Library to Implement Multi-tenant databases',
    url="https://github.com/citusdata/django-multitenant",
    author='Louise Grandjonc',
    author_email='louise.grandjonc@microsoft.com',

    # Classifiers help users find your project by categorizing it.
    #
    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        "Development Status :: 4 - Beta",
        "Topic :: Database",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers"
    ],

    keywords=("citus django multi tenant"
              "django postgres multi-tenant"),
    packages=find_packages(exclude=['tests']),  # Required
)



# Generated by Django 2.1.1 on 2019-03-01 08:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0004_auto_20190301_0834'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tenantnotidmodel',
            name='tenant_column',
            field=models.IntegerField(editable=False, primary_key=True, serialize=False),
        ),
    ]



# Generated by Django 2.2.9 on 2020-02-04 09:29

from django.db import migrations
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0016_auto_20191025_0844'),
    ]

    operations = [
        migrations.AddField(
            model_name='task',
            name='parent',
            field=django_multitenant.fields.TenantForeignKey(blank=True, db_index=False, null=True, on_delete=django.db.models.deletion.CASCADE, to='tests.Task'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.20 on 2020-04-12 06:03
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields
import django_multitenant.mixins

from django_multitenant.db import migrations as tenant_migrations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0022_merge_20200211_1000'),
    ]

    operations = [
        migrations.CreateModel(
            name='Revenue',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.CharField(max_length=30)),
                ('acc', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='revenues', to='tests.Account')),
                ('project', django_multitenant.fields.TenantForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='revenues', to='tests.Project')),
            ],
            options={
                'abstract': False,
            },
            bases=(django_multitenant.mixins.TenantModelMixin, models.Model),
        ),
        migrations.RunSQL("ALTER TABLE tests_revenue DROP CONSTRAINT tests_revenue_pkey CASCADE;"),
        tenant_migrations.Distribute('Revenue'),
        migrations.RunSQL("ALTER TABLE tests_revenue ADD CONSTRAINT tests_revenue_pkey PRIMARY KEY (acc_id, id);")
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-05-17 15:14
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
from django.conf import settings
import django_multitenant.fields
import django_multitenant.mixins

from django_multitenant.db import migrations as tenant_migrations


def get_operations():
    operations = [
        migrations.RunSQL("ALTER TABLE tests_tempmodel DROP CONSTRAINT tests_tempmodel_pkey CASCADE;"),
        migrations.RunSQL("ALTER TABLE tests_tempmodel ADD CONSTRAINT tests_tempmodel_pkey PRIMARY KEY (account_id, id);")]

    if settings.USE_CITUS:
        operations += [
            tenant_migrations.Distribute('TempModel'),
        ]

    return operations


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0010_auto_20190517_1514'),
    ]
    operations = get_operations()



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-05-17 16:06
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0011_distribute_new_table'),
    ]

    operations = [
        migrations.AlterField(
            model_name='projectmanager',
            name='project',
            field=django_multitenant.fields.TenantForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Project'),
        ),

        migrations.AddField(
            model_name='tempmodel',
            name='project',
            field=django_multitenant.fields.TenantForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, to='tests.Project'),
        ),
        migrations.AlterField(
            model_name='tempmodel',
            name='account',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='tests.Account', db_constraint=False),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11 on 2019-05-17 16:07
from __future__ import unicode_literals

from django.db import migrations
import django.db.models.deletion
import django_multitenant.fields


class Migration(migrations.Migration):

    dependencies = [
        ('tests', '0012_auto_20190517_1606'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='tempmodel',
            name='project',
        ),
        migrations.DeleteModel(
            name='TempModel',
        ),
    ]
